unit MenuToolBar;

interface

uses
  Windows, Messages, SysUtils, CommCtrl, Classes, Graphics, Controls, Forms, Dialogs,
  ToolWin, ComCtrls, ExtCtrls, Buttons, Menus;

type
  TIconButton = class(TToolButton)
    procedure Paint; override;
  end;

  TMenuToolBar = class(TToolBar)
  private
    { Private declarations }
    Panel: TPanel;
    MinimizeBtn,
    RestoreBtn,
    CloseBtn: TSpeedButton;
    IconButton: TIconButton;
    OldWndProc,
    WndHookProc: Pointer;
    OldClientProc,
    WndClientHookProc: Pointer;
    Form: TForm;
    Hooks: TList;
    procedure SetDefaultMenuFont;
  protected
    { Protected declarations }
    procedure WndProc(var Message: TMessage); override;
    procedure TBAddButtons(var Message: TMessage); message TB_ADDBUTTONS;
    procedure TBInsertButton(var Message: TMessage); message TB_INSERTBUTTON;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WndHook(var Message: TMessage); virtual;
    procedure WndClientHook(var Message: TMessage); virtual;
    procedure CloseBtnClick(Sender: TObject);
    procedure MinimizeBtnClick(Sender: TObject);
    procedure RestoreBtnClick(Sender: TObject);
    procedure ShowMDIElements;
    procedure HideMDIElements;
    procedure CreateMenuButtons;
  public
    { Public declarations }
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Font stored False;
  end;

procedure Register;

implementation

const
  TBSTYLE_AUTOSIZE = $0010;

var
  HookButton: TToolButton;
  Hook: HHOOK;

type
  THook = class(TObject)
  private
    Handle: THandle;
    Proc: Pointer;
    Menu: TMenuToolbar;
  protected
    procedure WndChildHook(var Message: TMessage); virtual;
  public
    constructor Create(AHandle: THandle; AMenu: TMenuToolbar);
    destructor Destroy; override;
  end;

constructor THook.Create(AHandle: THandle; AMenu: TMenuToolbar);
begin
  inherited Create;
  Handle:= AHandle;
  Menu:= AMenu;
  Proc:= Pointer(SetWindowLong(Handle, GWL_WNDPROC, Integer(MakeObjectInstance(WndChildHook))));
end;

destructor THook.Destroy;
begin
  FreeObjectInstance(Pointer(SetWindowLong(Handle, GWL_WNDPROC, Longint(Proc))));
  inherited;
end;

procedure TIconButton.Paint;
var
  MTB: TMenuToolbar;
  Icon: THandle;
begin
  MTB:= Parent as TMenuToolbar;
  Canvas.Brush.Color:= MTB.Color;
  Canvas.FillRect(Canvas.ClipRect);
  if MTB.Form.MDIChildCount>0 then
    begin
      Icon:= MTB.Form.ActiveMDIChild.Icon.Handle;
      if Icon=0 then
        Icon:= LoadIcon(HInstance, 'MAINICON');
      DrawIconEx(Canvas.Handle, 0, (Height-16) div 2, Icon, 16, 16, 0, 0, DI_NORMAL);
    end;
end;

function HookProc(nCode: Integer; wParam: Longint; var Msg: TMsg): Longint; stdcall;
var
  Pt: TPoint;
  TB: TToolButton;
  R: TRect;
  Control: TControl;
  Toolbar: TMenuToolbar;
  FForm: TForm;
  I: Integer;
begin
  if (nCode = MSGF_MENU) and Assigned(HookButton) then
    case Msg.Message of
    WM_MOUSEMOVE:
      begin
        Pt:= SmallPointToPoint(TSmallPoint(Msg.lParam));
        Toolbar:= HookButton.Parent as TMenuToolbar;
        Pt:= Toolbar.ScreenToClient(Pt);
        if Assigned(Toolbar) and PtInRect(Toolbar.ClientRect, Pt) then
          begin
            Control:= Toolbar.ControlAtPos(Pt, False);
            if (Control is TToolButton) then
              begin
                TB:= Control as TToolButton;
                if (TB<>HookButton) then
                  PostMessage(Toolbar.Handle, WM_LBUTTONDOWN, MK_LBUTTON, MakeLong(Pt.x, Pt.y));
              end
            else
              PostMessage(Toolbar.Handle, WM_LBUTTONDOWN, MK_LBUTTON, MakeLong(Pt.x, Pt.y));
          end;
      end;
    WM_LBUTTONDOWN,
    WM_RBUTTONDOWN:
      begin
        Toolbar:= HookButton.Parent as TMenuToolbar;
        Pt:= SmallPointToPoint(TSmallPoint(Msg.lParam));
        Control:= Toolbar.ControlAtPos(Pt, False);
        Result:= CallNextHookEx(Hook, nCode, wParam, Longint(@Msg));
        UnHookWindowsHookEx(Hook);
        Hook:= 0;
        HookButton:= nil;
        Exit;
      end;
    end;
  if Hook<>0 then
    Result:= CallNextHookEx(Hook, nCode, wParam, Longint(@Msg));
end;

constructor TMenuToolbar.Create(Owner: TComponent);
var
  I: Integer;
  TB: TTBButton;
  R: TRect;
  Component: TComponent;
begin
  inherited Create(Owner);
  EdgeBorders:= [];
  ShowCaptions:= True;
  Flat:= True;
  Height:= GetSystemMetrics(SM_CYMENU);
  AutoSize:= True;
  SetDefaultMenuFont;
  Component:= Owner;
  while Assigned(Component) and not (Component is TForm) do
    Component:= Component.Owner;
  if not Assigned(Component) then
    Exit;
  Form:= Component as TForm;
  if not (csDesigning in ComponentState) then
    begin
      WndHookProc := MakeObjectInstance(WndHook);
      OldWndProc := Pointer(GetWindowLong(Form.Handle, GWL_WNDPROC));
      SetWindowLong(Form.Handle, GWL_WNDPROC, Longint(WndHookProc));
    end;
  if not (csDesigning in ComponentState) and (Form.FormStyle=fsMDIForm) then
    begin
      Hooks:= TList.Create;
      WndClientHookProc := MakeObjectInstance(WndClientHook);
      OldClientProc := Pointer(GetWindowLong(Form.ClientHandle, GWL_WNDPROC));
      SetWindowLong(Form.ClientHandle, GWL_WNDPROC, Longint(WndClientHookProc));
    end;
end;

destructor TMenuToolbar.Destroy;
var
  I: Integer;
begin
  if Assigned(Hooks) then
    begin
      for I:= Hooks.Count-1 downto 0 do
        THook(Hooks.Items[I]).Free;
      Hooks.Free;
    end;
  inherited;
end;

procedure TMenuToolbar.ShowMDIElements;
var
  NCM: TNonClientMetrics;
  R: TRect;
begin
  if not Assigned(IconButton) then
    begin
      IconButton:= TIconButton.Create(Self);
      IconButton.Width:= 16;
      IconButton.Style:= tbsDropDown;
      IconButton.Parent:= Self;
      IconButton.Left:= 0;
      IconButton.Tag:= GetSystemMenu(Form.ActiveMDIChild.Handle, False);
      IconButton.ImageIndex:= Form.Handle;
      Width:= Width-1;
      Width:= Width+1;
    end;
  if not Assigned(Panel) then
    begin
      NCM.cbSize:= sizeof(NCM);
      SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, Pointer(@NCM), 0);
      Panel:= TPanel.Create(Self);
      Panel.Caption:= '';
      Panel.BevelInner:= bvNone;
      Panel.BevelOuter:= bvNone;
      Perform(TB_GETITEMRECT, ButtonCount-1, Longint(@R));
      Panel.Parent:= Self;
      Panel.Width:= Width-R.Right;
      CloseBtn:= TSpeedButton.Create(Panel);
      with CloseBtn do
        begin
          Width:= NCM.iSmCaptionWidth-2;
          Height:= NCM.iCaptionHeight-3;
          Font.Name:= 'Marlett';
          Font.Size:= 6;
          Caption:= 'r';
          Left:= Panel.Width - Width;
          Top:= (Panel.Height-Height) div 2;
          OnClick:= CloseBtnClick;
        end;
      RestoreBtn:= TSpeedButton.Create(Panel);
      with RestoreBtn do
        begin
          Width:= NCM.iSmCaptionWidth-2;
          Height:= NCM.iCaptionHeight-3;
          Font.Name:= 'Marlett';
          Font.Size:= 6;
          Caption:= '2';
          Left:= Panel.Width - 2*Width-2;
          Top:= (Panel.Height-Height) div 2;
          OnClick:= RestoreBtnClick;
        end;
      MinimizeBtn:= TSpeedButton.Create(Panel);
      with MinimizeBtn do
        begin
          Width:= NCM.iSmCaptionWidth-2;
          Height:= NCM.iCaptionHeight-3;
          Font.Name:= 'Marlett';
          Font.Size:= 6;
          Caption:= '0';
          Left:= Panel.Width - 3*Width-2;
          Top:= (Panel.Height-Height) div 2;
          OnClick:= MinimizeBtnClick;
        end;
      CloseBtn.Parent:= Panel;
      RestoreBtn.Parent:= Panel;
      MinimizeBtn.Parent:= Panel;
    end;
end;

procedure TMenuToolbar.HideMDIElements;
begin
  if Assigned(IconButton) then
    begin
      IconButton.Free;
      IconButton:= nil;
    end;
  if Assigned(Panel) then
    begin
      Panel.Free;
      Panel:= nil;
    end;
end;

procedure TMenuToolbar.CreateMenuButtons;
var
  TB: TToolButton;
  I, J: Integer;
  MII: TMenuItemInfo;
  Buffer: array[0..255]of Char;
begin
  if (csDesigning in ComponentState) or not Assigned(Form.Menu) then
    Exit;
  for I:= 0 to Form.Menu.Items.Count-1 do
    begin
      TB:= TToolButton.Create(Self);
      TB.Tag:= Form.Menu.Items[I].Handle;
      TB.Caption:= Form.Menu.Items[I].Caption;
      TB.OnClick:= Form.Menu.Items[I].OnClick;
      TB.ImageIndex:= Form.Menu.Items[I].GroupIndex+1;
      if Form.Menu.Items[I].Count>0 then
        TB.Style:= tbsDropDown
      else
        TB.Style:= tbsButton;
      TB.Parent:= Self;
    end;
  if Assigned(Form.ActiveMDIChild) and Assigned(Form.ActiveMDIChild.Menu) then
    for I:= 0 to Form.ActiveMDIChild.Menu.Items.Count-1 do
      begin
        TB:= TToolButton.Create(Self);
        TB.Tag:= Form.ActiveMDIChild.Menu.Items[I].Handle;
        TB.Caption:= Form.ActiveMDIChild.Menu.Items[I].Caption;
        TB.OnClick:= Form.ActiveMDIChild.Menu.Items[I].OnClick;
        TB.ImageIndex:= Form.ActiveMDIChild.Menu.Items[I].GroupIndex+1;
        if Form.ActiveMDIChild.Menu.Items[I].Count>0 then
          TB.Style:= tbsDropDown
        else
          TB.Style:= tbsButton;
        TB.Parent:= Self;
        for J:= 0 to ButtonCount-2 do
          if Buttons[J].ImageIndex>Abs(TB.ImageIndex) then
            begin
              TB.Left:= Buttons[J].Left;
              TB.ImageIndex:= -TB.ImageIndex;
              Break;
            end;
      end;
  for I:= 0 to ButtonCount-1 do
    if Buttons[I].ImageIndex>0 then
      Buttons[I].ImageIndex:= Form.Handle
    else
      Buttons[I].ImageIndex:= Form.ActiveMDIChild.Handle;
end;

procedure TMenuToolbar.WndProc(var Message: TMessage);
var
  TB: TControl;
  Pt: TPoint;
begin
  if not (csDesigning in ComponentState) then
    case Message.Msg of
    WM_LBUTTONDOWN:
      begin
        Pt:= SmallPointToPoint(TSmallPoint(Message.lParam));
        TB:= ControlAtPos(Pt, False);
        if (TB is TToolButton) then
          begin
            inherited;
            HookButton:= TB as TToolButton;
            if Assigned(HookButton.OnClick) then
              DefaultHandler(Message);
            if HookButton.Tag<>0 then
              begin
                if Hook=0 then
                  Hook:= SetWindowsHookEx(WH_MSGFILTER, @HookProc, 0, GetCurrentThreadID);
                Pt:= ClientToScreen(Point(HookButton.Left, HookButton.Height));
                TrackPopupMenuEx(HookButton.Tag, 0, Pt.x, Pt.y, HookButton.ImageIndex, nil);
                Message.Msg:= WM_LBUTTONUP;
                DefaultHandler(Message);
              end;
          end;
      end;
    WM_RBUTTONDOWN:
      begin
        Pt:= SmallPointToPoint(TSmallPoint(Message.lParam));
        TB:= ControlAtPos(Pt, False);
        if (TB is TToolButton) then
          begin
            inherited;
            HookButton:= TB as TToolButton;
            if Assigned(HookButton.OnClick) then
              DefaultHandler(Message);
            if HookButton.Tag<>0 then
              begin
                if Hook=0 then
                  Hook:= SetWindowsHookEx(WH_MSGFILTER, @HookProc, 0, GetCurrentThreadID);
                Pt:= ClientToScreen(Pt);
                TrackPopupMenuEx(HookButton.Tag, 0, Pt.x, Pt.y, HookButton.ImageIndex, nil);
                Message.Msg:= WM_LBUTTONUP;
                DefaultHandler(Message);
              end;
          end;
      end;
    WM_LBUTTONDBLCLK:
      begin
        Pt:= SmallPointToPoint(TSmallPoint(Message.lParam));
        TB:= ControlAtPos(Pt, False);
        if (TB is TIconButton) then
          begin
            inherited;
            if Form.MDIChildCount>0 then
              Form.ActiveMDIChild.Close;
          end;
      end;
    end;
  inherited;
end;

procedure TMenuToolbar.TBAddButtons(var Message: TMessage);
var
  I: Integer;
  TB: PTBBUTTON;
begin
  TB:= PTBButton(Message.LParam);
  for I:= 0 to Message.WParam-1 do
    begin
      if (TB^.fsStyle and TBSTYLE_SEP)=0 then
        TB^.fsStyle:= TB^.fsStyle or TBSTYLE_AUTOSIZE;
      Inc(TB, SizeOf(TTBBUTTON));
    end;
  DefaultHandler(Message);
end;

procedure TMenuToolbar.TBInsertButton(var Message: TMessage);
begin
  with PTBButton(Message.LParam)^ do
    if (fsStyle and TBSTYLE_SEP)=0 then
      fsStyle:= fsStyle or TBSTYLE_AUTOSIZE;
  DefaultHandler(Message);
end;

procedure TMenuToolbar.WMSize(var Message: TWMSize);
begin
  inherited;
  if Assigned(Panel) then
    begin
      Panel.Width:= Width-Panel.Left;
      if Assigned(CloseBtn) then
        begin
          CloseBtn.Left:= Panel.Width - CloseBtn.Width-Indent;
          RestoreBtn.Left:= Panel.Width - 2*CloseBtn.Width-2-Indent;
          MinimizeBtn.Left:= Panel.Width - 3*CloseBtn.Width-2-Indent;
        end;
    end;
end;

procedure TMenuToolbar.SetDefaultMenuFont;
var
  NCM: TNonClientMetrics;
begin
  NCM.cbSize:= sizeof(NCM);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, Pointer(@NCM), 0);
  with ( NCM.lfMenuFont ) do begin
    Font.Height:= lfHeight;
    if ( lfWeight > 500 ) then
      Font.Style:= [fsBold];
    if ( lfItalic <> 0 ) then
      Font.Style:= Font.Style + [fsItalic];
    if ( lfUnderline <> 0 ) then
      Font.Style:= Font.Style + [fsUnderline];
    if ( lfStrikeOut <> 0 ) then
      Font.Style:= Font.Style + [fsStrikeOut];
    Font.CharSet:= TFontCharSet(lfCharSet);
    Font.Pitch:= TFontPitch(lfPitchAndFamily);
    Font.Name:= lfFaceName;
  end;
end;

procedure TMenuToolbar.CloseBtnClick(Sender: TObject);
begin
  if Form.MDIChildCount>0 then
    Form.ActiveMDIChild.Close;
end;

procedure TMenuToolbar.MinimizeBtnClick(Sender: TObject);
begin
  if Form.MDIChildCount>0 then
    Form.ActiveMDIChild.WindowState:= wsMinimized;
end;

procedure TMenuToolbar.RestoreBtnClick(Sender: TObject);
begin
  if Form.MDIChildCount>0 then
    Form.ActiveMDIChild.WindowState:= wsNormal;
end;

procedure TMenuToolbar.WndHook(var Message: TMessage);
var
  R: TRect;
begin
  case Message.Msg of
  WM_NCCALCSIZE:
    begin
      Message.Result:= CallWindowProc(OldWndProc, Form.Handle, Message.Msg, Message.WParam, Message.LParam);
      R:= TWMNCCalcSize(Message).CalcSize_Params.rgrc[0];
      OffsetRect(R, 0, -GetSystemMetrics(SM_CYMENU));
      R.Bottom:= R.Bottom + GetSystemMetrics(SM_CYMENU);
      TWMNCCalcSize(Message).CalcSize_Params.rgrc[0]:= R;
      Exit;
    end;
  WM_NCHITTEST:
    begin
      Message.Result:= CallWindowProc(OldWndProc, Form.Handle, Message.Msg, Message.WParam, Message.LParam);
      if (Message.Result = HTMENU ) then
        Message.Result:= HTTRANSPARENT;
      Exit;
    end;
  WM_ACTIVATE:
    if ButtonCount=0 then
      CreateMenuButtons;
  WM_DESTROY:
    begin
      SetWindowLong(Form.Handle, GWL_WNDPROC, Longint(OldWndProc));
      FreeObjectInstance(WndHookProc);
    end;
  end;
  Message.Result:= CallWindowProc(OldWndProc, Form.Handle, Message.Msg, Message.WParam, Message.LParam);
end;

procedure TMenuToolbar.WndClientHook(var Message: TMessage);
var
  I: Integer;
begin
  case Message.Msg of
  WM_MDIMAXIMIZE:
    if Assigned(Panel) then
      ShowMDIElements;
  WM_MDIRESTORE, WM_MDICASCADE, WM_MDITILE:
    if Assigned(Panel) then
      HideMDIElements;
  WM_MDICREATE:
    begin
      Message.Result:= CallWindowProc(OldClientProc, Form.ClientHandle, Message.Msg, Message.WParam, Message.LParam);
      Hooks.Add(THook.Create(Message.Result, Self));
      Exit;
    end;
  WM_MDIDESTROY:
    if Assigned(Panel) and (Form.MDIChildCount=1) then
      HideMDIElements;
  WM_DESTROY:
    begin
      SetWindowLong(Form.ClientHandle, GWL_WNDPROC, Longint(OldClientProc));
      FreeObjectInstance(WndClientHookProc);
    end;
  end;
  Message.Result:= CallWindowProc(OldClientProc, Form.ClientHandle, Message.Msg, Message.WParam, Message.LParam);
end;

procedure THook.WndChildHook(var Message: TMessage);
var
  I: Integer;
  B: Boolean;
begin
  case Message.Msg of
  WM_SIZE:
    begin
      Message.Result:= CallWindowProc(Proc, Handle, Message.Msg, Message.WParam, Message.LParam);
      if TWMSize(Message).SizeType=SIZE_MAXIMIZED then
        Menu.ShowMDIElements
      else
        for I:= 0 to Menu.Form.MDIChildCount-1 do
          if Menu.Form.MDIChildren[I].Handle=Handle then
            begin
              if Menu.Form.MDIChildren[I].WindowState=wsMaximized then
                Menu.ShowMDIElements
              else
                Menu.HideMDIElements;
              Break;
            end;
        Exit;
    end;
  WM_MDIACTIVATE:
    begin
      Message.Result:= CallWindowProc(Proc, Handle, Message.Msg, Message.WParam, Message.LParam);
      with Menu do
        begin
          Perform(WM_SETREDRAW, 0, 0);
          B:= Assigned(Panel);
          if B then
            HideMDIElements;
          for I:= Menu.ButtonCount-1 downto 0 do
            Menu.Buttons[I].Free;
          Menu.CreateMenuButtons;
          if B then
            ShowMDIElements;
          Perform(WM_SETREDRAW, 1, 0);
        end;
      Exit;
    end;
  WM_DESTROY:
    begin
      Menu.Hooks.Delete(Menu.Hooks.IndexOf(Self));
      Message.Result:= CallWindowProc(Proc, Handle, Message.Msg, Message.WParam, Message.LParam);
      Free;
      Exit;
    end;
  end;
  Message.Result:= CallWindowProc(Proc, Handle, Message.Msg, Message.WParam, Message.LParam);
end;

procedure Register;
begin
  RegisterComponents('Win32', [TMenuToolbar]);
end;

initialization
  Hook:= 0;

end.
